Unit VGAScrl;

Interface

Procedure UseBitmap(TheBitmap: pointer; BitmapWidth: word);
Procedure ScrollUp(x1:Word;y1:Byte;x2:Word;y2,pixelstep:Byte);
Procedure ScrollDown(x1:Word;y1:Byte;x2:Word;y2,pixelstep:Byte);
Procedure ScrollLeft(x1:Word;y1:Byte;x2:Word;y2,pixelstep:Byte);
Procedure ScrollRight(x1:Word;y1:Byte;x2:Word;y2,pixelstep:Byte);

Implementation

Var CurrentBitmapSegment,
    CurrentBitmapOffset,
    CurrentBitmapWidth : word;


{ AX = X, BX = Y

  On exit: BX = Offset

  And there's no validation so make sure AX & BX are set OK!!
}

Procedure CalculateOffset; Assembler;
Asm
   PUSH DX
   PUSH AX
   MOV AX,CurrentBitmapWidth    { Make sure it's <= 320 }
   MUL BX
   MOV BX,AX
   POP AX
   POP DX
   ADD BX,AX
   ADD BX,CurrentBitmapOffset
End;






{
POST INCREMENTAL MOVE
---------------------

You could put this in MEMORY.SWG if you like as it uses 32 bit
instructions to copy data from A to B.

Expects:     DS:SI = Source
             ES:DI = Dest
             CX    = Count of bytes

Returns:     Zilch

Affects :    SI,DI, Direction Flag
}



Procedure PostIncMove; Assembler;
Asm
   CLD

   CMP CX,4             { # of bytes to move > 4 ? }
   JB @NoLongs

   PUSH CX
   SHR CX,2             { Divide count by 4 to give number of }
   REPZ                 { long words to move in one go }
   DB $66; MOVSW
   POP CX

@NoLongs:
   TEST CL,2            { Any words to move ? }
   JZ @NoWords
   MOVSW

@NoWords:
   TEST CL,1            { Or even a Byte ? }
   JZ @NoBytes
   MOVSB

@NoBytes:
End;








{
PRE DECREMENTAL MOVE
--------------------

Used for scrolling RIGHT. And this routine was a BAST to write!
It took AGES to debug!


Expects:     DS:SI = Source
             ES:DI = Dest
             CX    = Count of bytes

Returns:     Zilch

Affects :    SI,DI, Direction flag


}



Procedure PreDecMove; Assembler;
Asm
   STD

   CMP CX,4                     { Any long words to move ? }
   JB @CouldBeWord              { No, but there probably is at least 2 }

   PUSH CX                      { As CX is gonna be corrupted, better }
                                { push it on stack eh ? }
   SHR CX,2                     { Get number of long words to move -> CX }

   SUB SI,3                     { A MOVSD overwrites 4 bytes so this }
   SUB DI,3                     { must be taken into account }

@MoveLong:
   DB $66; MOVSW
   DEC CX
   JNZ @MoveLong
   ADD SI,3                     { Point to previous bytes written -1 }
   ADD DI,3

   POP CX

@CouldBeWord:
   TEST CL,2
   JZ @CouldBeByte

   DEC SI
   DEC DI
   MOVSW
   INC SI
   INC DI


@CouldBeByte:
   TEST CL,1                    { Is there a Byte left ? }
   JZ @NoBytes

   MOVSB

@NoBytes:
End;






{
Select bitmap to work with.

Passing in Ptr($a000,0) uses the VGA bitmap, for example.
BitmapWidth specifies width of bitmap in bytes. This can be any value
although it's standard to use 320.
}


Procedure UseBitmap(TheBitmap: pointer; BitmapWidth: word); Assembler;
Asm
   LES DI, TheBitmap
   MOV CurrentBitmapSegment,ES
   MOV CurrentBitmapOffset,DI
   MOV AX,BitmapWidth
   MOV CurrentBitmapWidth,AX
End;






{
Scroll a rectangular region of the Current Bitmap UP

Expects:  x1,y1,x2,y2 define the rectangular area.
          pixelstep is the number of pixels UP you wish to scroll.

Returns:  Nothing

Affects : AX,BX,CX,DX,SI,DI,ES,Direction flag
}



Procedure ScrollUp(x1:Word;y1:Byte;x2:Word;y2,pixelstep:Byte); Assembler;
Asm
     MOV AX,X1
     XOR BH,BH
     MOV BL,Y1
     MOV DX,BX               { Instead of pushing BX on stack }
     CALL CalculateOffset
     MOV SI,BX

     MOV BX,DX
     XOR CH,CH
     MOV CL,pixelstep
     SUB BX,CX
     CALL CalculateOffset    { Calculate address for X1,(Y1 - pixelstep) }
     MOV DI,BX

     PUSH DS

     MOV AX,CurrentBitmapSegment
     MOV DS,AX
     MOV ES,AX

     MOV CX,X2
     SUB CX,X1
     INC CX             { CX = Number of bytes per line to move }

     MOV DL,Y2
     SUB DL,Y1
     INC DL             { DL = Number of lines to move }

     PUSH BP
     MOV BP,SI          { Oh well that's knackered local var access then :) }
     MOV BX,DI


@Outer:
     MOV AX,320
     ADD BP,AX
     ADD BX,AX

     CALL PostIncMove

     MOV SI,BP          { point source & dest to next lines }
     MOV DI,BX

     DEC DL
     JNZ @Outer

     POP BP             { Restore the 2 essential registers }
     POP DS
End;







{
Scroll a rectangular region of the Current Bitmap DOWN

Expects:  x1,y1,x2,y2 define the rectangular area.
          pixelstep is the number of pixels down you wish to scroll.

Returns:  Nothing

Affects : AX,BX,CX,DX,SI,DI,ES, Direction flag
}



Procedure ScrollDown(x1:Word;y1:Byte;x2:Word;y2,pixelstep:Byte); Assembler;
Asm
     MOV AX,X1
     XOR BH,BH
     MOV BL,Y2
     MOV DX,BX                { Instead of pushing BX on stack }
     CALL CalculateOffset
     MOV SI,BX

     MOV BX,DX
     XOR CH,CH
     MOV CL,pixelstep
     ADD BX,CX
     CALL CalculateOffset     { Calculate address for X1,(Y1 + pixelstep) }
     MOV DI,BX

     PUSH DS

     MOV AX,CurrentBitmapSegment
     MOV DS,AX
     MOV ES,AX

     MOV CX,X2
     SUB CX,X1
     INC CX             { CX = Number of bytes per line to move }

     MOV DL,Y2
     SUB DL,Y1
     INC DL             { DL = Number of lines to move }


     PUSH BP
     MOV BP,SI          { Oh well that's knackered local var access then :) }
     MOV BX,DI          { ^ Do you get this deja vu feeling I've said that
                          before? :) }


@Outer:
     MOV AX,320
     SUB BP,AX
     SUB BX,AX

     CALL PostIncMove

     MOV SI,BP          { point source & dest to next lines }
     MOV DI,BX

     DEC DL
     JNZ @Outer

     POP BP             { Restore the 2 essential registers }
     POP DS
End;








{
Scroll a rectangular region of the Current Bitmap LEFT

Expects:  x1,y1,x2,y2 define the rectangular area.
          pixelstep is the number of pixels left you wish to scroll.

Returns:  Nothing

Affects : AX,BX,CX,DX,SI,DI,ES, Direction flag
}



Procedure ScrollLeft(x1:Word;y1:Byte;x2:Word;y2,pixelstep:Byte); Assembler;
Asm
     MOV AX,X1
     XOR BH,BH
     MOV BL,Y1
     MOV DX,BX                { Instead of pushing BX on stack }
     CALL CalculateOffset
     MOV SI,BX

     MOV BX,DX
     XOR CH,CH
     MOV CL,pixelstep
     SUB AX,CX
     CALL CalculateOffset    { Calculate address for (X1-pixelstep),Y1 }
     MOV DI,BX

     PUSH DS
     MOV AX,CurrentBitmapSegment
     MOV DS,AX
     MOV ES,AX

     MOV CX,X2
     SUB CX,X1
     INC CX                  { CX = Number of bytes per line to move }

     MOV DL,Y2
     SUB DL,Y1
     INC DL                  { DL = Number of lines to move }

     PUSH BP
     MOV BP,SI               { Oh well that's knackered local var
                               access then (spot the block copying) :) }
     MOV BX,DI


@Outer:
     MOV AX,320              { Bytes per scan line }

     ADD BP,AX
     ADD BX,AX

     CALL PostIncMove

     MOV SI,BP               { point source & dest to next lines }
     MOV DI,BX

     DEC DL
     JNZ @Outer

     POP BP                  { Restore the 2 essential registers }
     POP DS
End;







{
Scroll a rectangular region of the Current Bitmap RIGHT

Expects:  x1,y1,x2,y2 define the rectangular area.
          pixelstep is the number of pixels right you wish to scroll.

Returns:  Nothing

Affects : AX,BX,CX,DX,SI,DI,ES, Direction flag

Notes:    The PreInc routine was a bloody nightmare to write so it
          had better be appreciated :)
}



Procedure ScrollRight(x1:Word;y1:Byte;x2:Word;y2,pixelstep:Byte); Assembler;
Asm
     MOV AX,X2
     XOR BH,BH
     MOV BL,Y1
     MOV DX,BX                { Instead of pushing BX on stack }
     CALL CalculateOffset
     MOV SI,BX

     MOV BX,DX
     XOR CH,CH
     MOV CL,pixelstep
     ADD AX,CX
     CALL CalculateOffset    { Calculate address for (X1-pixelstep),Y1 }
     MOV DI,BX

     PUSH DS

     MOV AX,CurrentBitmapSegment
     MOV DS,AX
     MOV ES,AX

     MOV CX,X2
     SUB CX,X1
     INC CX             { CX = Number of bytes per line to move }

     MOV DL,Y2
     SUB DL,Y1
     INC DL             { DL = Number of lines to move }


     PUSH BP
     MOV BP,SI          { Oh well that's knackered local var access then :) }
     MOV BX,DI

@Outer:
     MOV AX,320
     ADD BP,AX
     ADD BX,AX

     CALL PreDecMove

     MOV SI,BP
     MOV DI,BX
     DEC DL
     JNZ @Outer

     POP BP
     POP DS
End;


Begin
     UseBitmap(Ptr($a000,0),320);
End.